/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * gwin.c: gtk window management routines for pho, an image viewer.
 *
 * Copyright 2004,2009 by Akkana Peck.
 * You are free to use or modify this code under the Gnu Public License.
 */

#include "pho.h"
#include "dialogs.h"
#include "exif/phoexif.h"

#include <stdio.h>
#include <stdlib.h>       /* for exit() */
#include <string.h>
#include <unistd.h>

#include <gdk/gdk.h>

/* Some window managers don't deal well with windows that resize,
 * or don't retain focus if a resized window no longer contains
 * the mouse pointer.
 * Offer an option to make new windows instead.
 */
int gMakeNewWindows = 0;

/* The size our window frame adds on the top left of the image */
gint sFrameWidth = 10;
gint sFrameHeight = 28;
int sHaveFrameSize = 0;

gint gPhysMonitorWidth = 0;
gint gPhysMonitorHeight = 0;

/* Initialize the "black" color */
static GdkColor sBlack = { 0x0000, 0x0000, 0x0000 };


/* gtk related window attributes */
GtkWidget *gWin = 0;
static GtkWidget *sDrawingArea = 0;

static void NewWindow(); /* forward */

static void hide_cursor(GtkWidget* w)
{
    static char invisible_cursor_bits[] = { 0x0 };
    static GdkBitmap *empty_bitmap = 0;
    static GdkCursor* cursor = 0;

    if (empty_bitmap == 0 || cursor == 0) {
        empty_bitmap = gdk_bitmap_create_from_data(NULL,
                                                   invisible_cursor_bits,
                                                   1, 1);
        cursor = gdk_cursor_new_from_pixmap (empty_bitmap, empty_bitmap, &sBlack, &sBlack, 1, 1);
    }
    gdk_window_set_cursor(w->window, cursor);

    /* If we need to free this, do it this way:
    gdk_cursor_unref(cursor);
    g_object_unref(empty_bitmap);
     */
}

static void show_cursor(GtkWidget* w)
{
    gdk_window_set_cursor(w->window, NULL);
}    

/*
 * Adjust gMonitorWidth/Height and sFrameWidth/Height
 * according to the display mode and current frames,
 * then resize the image accordingly.
 * Call this when the image size changes --
 * zoom, shift in/out of presentation mode, etc.
 */
static void AdjustScreenSize()
{
    if (gDisplayMode == PHO_DISPLAY_PRESENTATION) {
        gMonitorWidth = gPhysMonitorWidth;
        gMonitorHeight = gPhysMonitorHeight;
    }
    else {
        gint width, height;
        GdkRectangle rect;
        gdk_drawable_get_size(gWin->window, &width, &height);
        gdk_window_get_frame_extents(gWin->window, &rect);
        sFrameWidth = rect.width - width;
        sFrameHeight = rect.height - height;
        if (gDebug)
            printf("Frame size: %d, %d\n", sFrameWidth, sFrameHeight);

        gMonitorWidth = gPhysMonitorWidth - sFrameWidth;
        gMonitorHeight = gPhysMonitorHeight - sFrameHeight;
    }

    ScaleAndRotate(gCurImage, 0);
}

static void CenterWindow(GtkWidget* win)
{
    gint w, h;
    gtk_window_get_size(GTK_WINDOW(gWin), &w, &h);
    gtk_window_set_gravity(GTK_WINDOW(win), GDK_GRAVITY_CENTER);
    gtk_window_move(GTK_WINDOW(win),
                    (gMonitorWidth - gCurImage->curWidth)/2,
                    (gMonitorHeight - gCurImage->curHeight)/2);
    printf("Set gravity center and moved to (%d, %d) (%d-%d, %d-%d)\n",
           (gMonitorWidth - gCurImage->curWidth)/2,
           (gMonitorHeight - gCurImage->curHeight)/2,
           gMonitorWidth, gCurImage->curWidth,
           gMonitorHeight, gCurImage->curHeight);
    gtk_window_set_gravity(GTK_WINDOW(win), GDK_GRAVITY_NORTH_WEST);
}

void SetViewModes(int dispmode, int scalemode, double scalefactor)
{
    if (dispmode == gDisplayMode && scalemode == gScaleMode
        && scalefactor == gScaleRatio)
        return;
    gDisplayMode = dispmode;
    gScaleMode = scalemode;
    gScaleRatio = scalefactor;

    if (gWin && sDrawingArea) {    /* Changing an existing window */
        if (dispmode == PHO_DISPLAY_PRESENTATION) {
            printf("Changing an existing window back to presentation mode\n");
            hide_cursor(sDrawingArea);

            gtk_window_fullscreen(GTK_WINDOW(gWin));
            gtk_window_move(GTK_WINDOW(gWin), 0, 0);

            //AdjustScreenSize();
        }
        else {
            printf("Changing an existing window back to normal mode\n");
            printf("Monitor size is currently %dx%d\n",
                   gMonitorWidth, gMonitorHeight);
            show_cursor(sDrawingArea);
            gtk_window_unfullscreen(GTK_WINDOW(gWin));
            //AdjustScreenSize();
            printf("Monitor size after AdjustScreenSize: %dx%d\n",
                   gMonitorWidth, gMonitorHeight);
            CenterWindow(gWin);
        }
    }
    if (dispmode == PHO_DISPLAY_KEYWORDS) {
        if (gWin)
            ShowKeywordsDialog();
        gScaleMode = PHO_SCALE_SCREEN_RATIO;
        gScaleRatio = .5;
    }
    else {
        if (gWin)
            HideKeywordsDialog();
    }
    printf("End of SetDisplayMode: gDisplayMode = %d\n", gDisplayMode);
}

/* DrawImage is called from the expose callback.
 * It assumes we already have the image in gImage.
 */
void DrawImage()
{
    int dstX = 0, dstY = 0;
    char title[BUFSIZ];
#   define TITLELEN ((sizeof title) / (sizeof *title))

    if (gDebug) {
        printf("DrawImage %s, %dx%d\n", gCurImage->filename,
               gCurImage->curWidth, gCurImage->curHeight);
    }
    if (gImage == 0 || gWin == 0 || sDrawingArea == 0) return;

    if (gDisplayMode == PHO_DISPLAY_PRESENTATION) {
        gint width, height;
        gdk_window_clear(sDrawingArea->window);

        /* Center the image. This has to be done according to
         * the current window size, not the phys monitor size,
         * because in the xinerama case, gtk_window_fullscreen()
         * only fullscreens the current monitor, not all of them.
         */
        gtk_window_get_size(GTK_WINDOW(gWin), &width, &height);
        dstX = (width - gCurImage->curWidth) / 2;
        dstY = (height - gCurImage->curHeight) / 2;
    }
    else {
        /* Update the titlebar */
        sprintf(title, "pho: %s (%d x %d)", gCurImage->filename,
                gCurImage->trueWidth, gCurImage->trueHeight);
        if (HasExif())
        {
            const char* date = ExifGetString(ExifDate);
            if (date && date[0]) {
                /* Make sure there's room */
                if (strlen(title) + strlen(date) + 3 < TITLELEN)
                    strcat(title, " (");
                strcat(title, date);
                strcat(title, ")");
            }
        }
        if (gScaleMode == PHO_SCALE_FULLSIZE)
            strcat(title, " (fullsize)");
        else if (gScaleMode == PHO_SCALE_FULLSCREEN)
            strcat(title, " (fullscreen)");
        else if (gScaleMode == PHO_SCALE_IMG_RATIO ||
                 gScaleMode == PHO_SCALE_SCREEN_RATIO) {
            char* titleEnd = title + strlen(title);
            if (gScaleRatio < 1)
                sprintf(titleEnd, " [%s/ %d]",
                        (gScaleMode == PHO_SCALE_IMG_RATIO ? "fullsize " : ""),
                        (int)(1. / gScaleRatio));
            else
                sprintf(titleEnd, " [%s* %d]",
                        (gScaleMode == PHO_SCALE_IMG_RATIO ? "fullsize " : ""),
                        (int)gScaleRatio);
        }
        gtk_window_set_title(GTK_WINDOW(gWin), title);

        if (gDisplayMode == PHO_DISPLAY_KEYWORDS)
            ShowKeywordsDialog(gCurImage);
    }

    gdk_pixbuf_render_to_drawable(gImage, sDrawingArea->window,
                   sDrawingArea->style->fg_gc[GTK_WIDGET_STATE(sDrawingArea)],
                                  0, 0, dstX, dstY,
                                  gCurImage->curWidth, gCurImage->curHeight,
                                  GDK_RGB_DITHER_NONE, 0, 0);

    UpdateInfoDialog(gCurImage);
}


/* An expose event has a GdkRectangle area and a GdkRegion *region
 * as well as gint count of subsequent expose events.
 * Unfortunately count is always 0.
 */
static gint HandleExpose(GtkWidget* widget, GdkEventExpose* event)
{
    gint width, height;
    printf("In Expose, window state is %d\n",
           gdk_window_get_state(gWin->window));
    gdk_drawable_get_size(widget->window, &width, &height);
    if (gDebug) {
        printf("HandleExpose: area %dx%d +%d+%d in window %dx%d\n",
               event->area.width, event->area.height,
               event->area.x, event->area.y,
               width, height);
        if (event->area.x != 0 || event->area.y != 0) {
            if (event->area.width != width || event->area.height != height)
                printf("*** Expose different from window size!\n");
            if (event->area.width != gCurImage->curWidth ||
                event->area.height != gCurImage->curHeight)
                printf("** Expose different from actual image size of %dx%d!\n",
                       gCurImage->curWidth, gCurImage->curHeight);
        }
    }

    /* Make sure the window can resize smaller */
    if (!gMakeNewWindows)
        gtk_widget_set_size_request(GTK_WIDGET(gWin), 1, 1);

    if (!sHaveFrameSize) {  // XXXXXXXXXXXXXXXXXXXXXXXXXXXXX
        int oldw = sFrameWidth;
        int oldh = sFrameHeight;
        sHaveFrameSize = 1;
        AdjustScreenSize();
        if (sFrameWidth != oldw || sFrameHeight != oldh) {
            gtk_window_resize(GTK_WINDOW(gWin),
                              gCurImage->curWidth, gCurImage->curHeight);
        }
        //MaybeMove();
    }

    DrawImage();

    return TRUE;
}

void EndSession()
{
    gCurImage = 0;
    UpdateInfoDialog();
    if (gDisplayMode == PHO_DISPLAY_KEYWORDS)
        ShowKeywordsDialog();

    PrintNotes();
    gtk_main_quit();
    /* This doesn't always quit!  So make sure: */
    exit(0);
}

static gint HandleDelete(GtkWidget* widget, GdkEventKey* event, gpointer data)
{
    /* If you return FALSE in the "delete_event" signal handler,
     * GTK will emit the "destroy" signal. Returning TRUE means
     * you don't want the window to be destroyed.
     * This is useful for popping up 'are you sure you want to quit?'
     * type dialogs. */
    EndSession();
    return TRUE;
}

/* If we  resized, see if we might move off the screen.
 * Make sure we don't do that, assuming the image is still
 * small enough to fit. Then request focus from the window manager.
 * Try to move the window as little as possible.
 */
static void MaybeMove()
{
    gint x, y, w, h, nx, ny;

    if (gCurImage->curWidth > gPhysMonitorWidth
        || gCurImage->curHeight > gPhysMonitorHeight)
        return;

    /* If we don't have a window yet, don't move it */
    if (!gWin)
        return;

    /* If we're in presentation mode, never move the window */
    if (gDisplayMode == PHO_DISPLAY_PRESENTATION
        || gDisplayMode == PHO_DISPLAY_KEYWORDS)
        return;

    gtk_window_get_position(GTK_WINDOW(gWin), &x, &y);
    nx = x;  ny = y;
    gtk_window_get_size(GTK_WINDOW(gWin), &w, &h);
    /* printf("Currently (%d, %d) %d x %d\n", x, y, w, h); */

    /* See if it would overflow off the screen */
    if (x + gCurImage->curWidth > gMonitorWidth)
        nx = gMonitorWidth - gCurImage->curWidth;
    if (nx < 0)
        nx = 0;
    if (y + gCurImage->curHeight > gMonitorHeight)
        ny = gMonitorHeight - gCurImage->curHeight;
    if (ny < 0)
        ny = 0;

    if (x != nx || y != ny) {
        gtk_window_move(GTK_WINDOW(gWin), nx, ny);
    }

    /* Request focus from the window manager.
     * This is pretty much a no-op, but what the heck.
     * However, it also messes with NewWindow's ability
     * to get the frame size because it maps the window before
     * we call gtk_widget_show().
    gtk_window_present(GTK_WINDOW(gWin));
     */
}

/* Make a new window, destroying the old one. */
static void NewWindow()
{
    gint root_x = -1;
    gint root_y = -1;
    printf("NewWindow()\n");

    if (gWin) {
        gdk_window_get_position(gWin->window, &root_x, &root_y);
        gtk_object_destroy(GTK_OBJECT(gWin));
    }

    gWin = gtk_window_new(GTK_WINDOW_TOPLEVEL);

    gtk_window_set_wmclass(GTK_WINDOW(gWin), "pho", "Pho");

    /* Window manager delete */
    gtk_signal_connect(GTK_OBJECT(gWin), "delete_event",
                       (GtkSignalFunc)HandleDelete, 0);

    /* This event occurs when we call gtk_widget_destroy() on the window,
     * or if we return FALSE in the "delete_event" callback.
    gtk_signal_connect(GTK_OBJECT(gWin), "destroy",
                       (GtkSignalFunc)HandleDestroy, 0);
     */

    /* KeyPress events on the drawing area don't come through --
     * they have to be on the window.
     */
    gtk_signal_connect(GTK_OBJECT(gWin), "key_press_event",
                       (GtkSignalFunc)HandleGlobalKeys, 0);

    sDrawingArea = gtk_drawing_area_new();
    gtk_container_add(GTK_CONTAINER(gWin), sDrawingArea);
    gtk_widget_show(sDrawingArea);
    /* This can't be done in expose: it causes one of those extra
     * spurious expose events that gtk so loves.
     */
    gtk_widget_modify_bg(sDrawingArea, GTK_STATE_NORMAL, &sBlack);

    if (gDisplayMode == PHO_DISPLAY_PRESENTATION) {
        gtk_drawing_area_size(GTK_DRAWING_AREA(sDrawingArea),
                              gPhysMonitorWidth, gPhysMonitorHeight);
        gtk_window_fullscreen(GTK_WINDOW(gWin));
    }
    else {
        gtk_drawing_area_size(GTK_DRAWING_AREA(sDrawingArea),
                              gCurImage->curWidth, gCurImage->curHeight);
        gtk_window_unfullscreen(GTK_WINDOW(gWin));
    }

#if 0
    /* This doesn't seem to make any difference. */
    gtk_widget_set_events(sDrawingArea,
                          GDK_EXPOSURE_MASK
                          | GDK_KEY_PRESS_MASK
                          | GDK_STRUCTURE_MASK );
#endif

    gtk_signal_connect(GTK_OBJECT(sDrawingArea), "expose_event",
                       (GtkSignalFunc)HandleExpose, 0);

#if 0
    /* Track in/out of fullscreen mode -- only works on toplevel windows */
    gtk_signal_connect (GTK_OBJECT(gWin), "configure_event",
                        (GtkSignalFunc)HandleExpose, NULL);
    gtk_signal_connect (GTK_OBJECT(gWin), "window_state_event",
		      (GtkSignalFunc)HandleWindowState, NULL);
#endif


    if (gDisplayMode == PHO_DISPLAY_PRESENTATION)
        hide_cursor(sDrawingArea);

    /* If not in presentation mode, and if this is the first time through,
     * get the frame size from the window manager (or from defaults)
     * and make a new window the right size.
     */
    else if (gCurImage->curWidth + sFrameWidth > gMonitorWidth
             || gCurImage->curHeight + sFrameHeight > gMonitorHeight) {
        /* Probably too big for framed window to fit on screen.
         * So maximize it. It would be best to maximize in only
         * one dimension, but gtk unfortunately doesn't allow that.
         */
#if 0
        static Atom maxvert = 0;
        static Atom maxhorz = 0;
        if (maxvert == 0)
            maxvert = gdk_atom_intern("_NET_WM_STATE_MAXIMIZED_VERT", FALSE);
        if (maxhorz == 0)
            maxhorz = gdk_atom_intern("_NET_WM_STATE_MAXIMIZED_HORZ", FALSE);
        gdk_wmspec_change_state (TRUE, window, maxvert, maxhorz);

        //printf("Before maximizing, window state is %d\n",
        //       gdk_window_get_state(gWin->window));
        gdk_window_maximize(gWin->window);
        //printf("After maximizing, window state is %d\n",
        //       gdk_window_get_state(gWin->window));
#endif
    }
    else
        /* Request a window position, based on the image size
         * and the position of the previous window.
         */
        MaybeMove();

    printf("Before show(): is window visible? %d, viewable? %d, mapped? %d\n",
           gdk_window_is_visible(gWin->window),
           gdk_window_is_viewable(gWin->window),
           GTK_WIDGET_MAPPED(gWin));
    printf("Widget flags: %p\n", (void*)GTK_WIDGET_FLAGS(gWin));
    printf("Showing window\n");
    gtk_widget_show(gWin);
    printf("After show(): is window visible? %d, viewable? %d, mapped? %d\n",
           gdk_window_is_visible(gWin->window),
           gdk_window_is_viewable(gWin->window),
           GTK_WIDGET_MAPPED(gWin));
    printf("Widget flags: %p\n", (void*)GTK_WIDGET_FLAGS(gWin));
}

/**
 * gdk_window_focus:
 * @window: a #GdkWindow
 * @timestamp: timestamp of the event triggering the window focus
 *
 * Sets keyboard focus to @window. If @window is not onscreen this
 * will not work. In most cases, gtk_window_present() should be used on
 * a #GtkWindow, rather than calling this function.
 *
 * For Pho: this is a replacement for gdk_window_focus
 * due to the issue in http://bugzilla.gnome.org/show_bug.cgi?id=150668
 * 
 **/
#define GDK_WINDOW_DISPLAY(win)       gdk_drawable_get_display(win)
#define GDK_WINDOW_SCREEN(win)	      gdk_drawable_get_screen(win)
#define GDK_WINDOW_XROOTWIN(win)      GDK_ROOT_WINDOW()

/* PrepareWindow is responsible for making the window the right
 * size and position, so the user doesn't see flickering.
 * It may actually make a new window, or it may just resize and/or
 * reposition the existing window.
 */
void PrepareWindow()
{
    if (gMakeNewWindows || gWin == 0) {
        NewWindow();
        return;
    }

    /* Otherwise, resize and reposition the current window. */

    if (gDisplayMode == PHO_DISPLAY_PRESENTATION) {
        /* XXX shouldn't have to do this every time */
        gtk_drawing_area_size(GTK_DRAWING_AREA(sDrawingArea),
                              gPhysMonitorWidth, gPhysMonitorHeight);
    }
    else {
        gint winwidth, winheight;

        gdk_drawable_get_size(gWin->window, &winwidth, &winheight);
        gdk_drawable_get_size(sDrawingArea->window, &winwidth, &winheight);

        /* We need to size the actual window, not just the drawing area.
         * Resizing the drawing area will resize the window for many
         * window managers, but not for metacity.
         *
         * Worse, metacity maximizes a window if the initial size is
         * bigger in either dimension than the screen size.
         * Since we can't be sure about the size of the wm decorations,
         * we will probably hit this and get unintentionally maximized,
         * after which metacity refuses to resize the window any smaller.
         * (Mac OS X apparently does this too.)
         * So force non-maximal mode.  (Users who want a maximized
         * window will probably prefer fullscreen mode anyway.)
         */
        gtk_window_unfullscreen(GTK_WINDOW(gWin));
        gtk_window_unmaximize(GTK_WINDOW(gWin));

        /* XXX Without the next line, we may get no expose events!
         * Likewise, if the next line doesn't actually resize anything
         * we may not get an expose event.
         */
        if (gCurImage->curWidth != winwidth
            || gCurImage->curHeight != winheight) {
            gtk_window_resize(GTK_WINDOW(gWin),
                              gCurImage->curWidth, gCurImage->curHeight);

            /* Unfortunately, on OS X this resize may not work,
             * if it puts part ofthe window off-screen; in which case
             * we won't get an Expose event. So if that happened,
             * force a redraw:
             */
            /* Just as unfortunately, get_size probably isn't reliable
             * on Linux either, until after the window manager has had
             * a chance to act on the resize request
             * (at which time we'll get another Configure notify).
             * So the dilemma is: we need to call DrawImage now in the
             * case where there will be no further events. But if there
             * are further events, we want to wait for them and not
             */
            gdk_drawable_get_size(sDrawingArea->window, &winwidth, &winheight);
            if (gCurImage->curWidth != winwidth
                || gCurImage->curHeight != winheight) {
                if (gDebug)
                    printf("Resize didn't work! Forcing redraw\n");
                DrawImage();
            }
        }

        /* If we didn't resize the window, then we won't get an expose
         * event, and hence DrawImage won't be called. So call it explicitly:
         */
        else {
            DrawImage();
        }

        /* Try to ensure that the window will be over the cursor
         * (so it will still have focus -- some window managers will
         * lose focus otherwise). But not in Keywords mode, where the
         * mouse should be over the Keywords dialog, not necessarily
         * the image window.
         */
        MaybeMove();
    }

    /* Request the focus.
     * Neither gtk_window_present nor gdk_window_focus seem to work.
     */
#ifdef TEST_FOCUS
    pho_window_focus(gWin->window, GDK_CURRENT_TIME);

    /* None of these actually work!  Is there any way to get keyboard
     * focus into a window?
     */
    gtk_window_activate_focus(GTK_WINDOW(gWin));
    gtk_window_present(GTK_WINDOW(gWin));

    GdkGrabStatus grabstat;
    printf("Trying desperately to grab focus!\n");
    gtk_window_present(GTK_WINDOW(gWin));
    gtk_widget_grab_focus(gWin);
    /*gdk_x11_display_grab(gdk_drawable_get_display(gWin->window));*/
    grabstat = gdk_keyboard_grab(gWin->window, FALSE, GDK_CURRENT_TIME);
    printf("Grab status: %d\n", grabstat);
    gdk_window_lower(gWin->window);
    /*gdk_window_raise(gWin->window);*/
    _gdk_x11_set_input_focus_safe(gdk_drawable_get_display(gWin->window),
                                  GDK_WINDOW_XID(gWin->window),
                                  TRUE,
                                  GDK_CURRENT_TIME);
#endif /* TEST_FOCUS */

    /* This will be redundant IF we get an expose event for the window;
     * but if we don't, then we need it here. How do we ensure that
     * we get exactly one expose event?
     * Though in practice, we always get at least one expose event,
     * and sometimes (if the window was resized) several.
     * So this call is probably never needed:
    DrawImage();
     */
}

